# from __future__ import print_function
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torchvision import datasets, transforms
from torch.autograd import Variable
from model import G12, G21
from data_loader import get_loader
import os
import pdb
import mnistm

# Training settings
parser = argparse.ArgumentParser(description='PyTorch MNIST Example')
parser.add_argument('--batch-size', type=int, default=64, metavar='N',
                    help='input batch size for training (default: 64)')
parser.add_argument('--test-batch-size', type=int, default=1000, metavar='N',
                    help='input batch size for testing (default: 1000)')
parser.add_argument('--epochs', type=int, default=30, metavar='N',
                    help='number of epochs to train (default: 10)')
parser.add_argument('--lr', type=float, default=0.01, metavar='LR',
                    help='learning rate (default: 0.01)')
parser.add_argument('--momentum', type=float, default=0.5, metavar='M',
                    help='SGD momentum (default: 0.5)')
parser.add_argument('--no-cuda', action='store_true', default=False,
                    help='disables CUDA training')
parser.add_argument('--seed', type=int, default=1, metavar='S',
                    help='random seed (default: 1)')
parser.add_argument('--log-interval', type=int, default=10, metavar='N',
                    help='how many batches to wait before logging training status')
parser.add_argument('--model', required=True, help='which pretrained model to load')
parser.add_argument('--iter', required=True, help='which iter of pretrained model to load')
# 0：s2m 1:u2m 2:m2u
parser.add_argument('--exp_id', type=int, default=0)
args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

gpu_ids = 0

torch.manual_seed(args.seed)
if args.cuda:
    torch.cuda.manual_seed(args.seed)

kwargs = {'num_workers': 1, 'pin_memory': True} if args.cuda else {}
train_loader = torch.utils.data.DataLoader(
    datasets.MNIST('./mnist', train=True, download=True,
                   transform=transforms.Compose([
                       transforms.Scale(32),
                       transforms.ToTensor(),
                       transforms.Normalize((0.5,), (0.5,))
                   ])),
    batch_size=args.batch_size, shuffle=True, **kwargs)
test_loader = torch.utils.data.DataLoader(
    datasets.MNIST('./mnist', train=False, transform=transforms.Compose([
        transforms.Scale(32),
        transforms.ToTensor(),
        transforms.Normalize((0.5,), (0.5,))
    ])),
    batch_size=1, shuffle=True, **kwargs)

transform = transforms.Compose([
    transforms.Scale(32),
    transforms.ToTensor(),
    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])
transform_mnist = transforms.Compose([
    transforms.Scale(32),
    transforms.ToTensor(),
    transforms.Normalize([0.5], [0.5])])
svhn_train = datasets.SVHN(root='./svhn', download=True, transform=transform, split='train')
svhn_test = datasets.SVHN(root='./svhn', download=True, transform=transform, split='test')
usps_test = datasets.USPS(root='./mnist', download=True, transform=transform_mnist, train=False)
usps_train = datasets.USPS(root='./mnist', download=True, transform=transform_mnist, train=True)
source_test = svhn_test
source_train = svhn_train
# if args.exp_id == 1:
#     source_test = usps_test
#     source_train = usps_train
svhn_test_loader = torch.utils.data.DataLoader(dataset=source_test,
                                        batch_size=1,
                                        shuffle=False,
                                        num_workers=2)

svhn_train_loader = torch.utils.data.DataLoader(dataset=source_train,
                                                batch_size=args.batch_size,
                                                shuffle=False,
                                                num_workers=2)
svhn_test_loader = torch.utils.data.DataLoader(dataset=source_test,
                                               batch_size=args.test_batch_size,
                                               shuffle=False,
                                               num_workers=2)
if args.exp_id == 1:
    svhn_test_loader = torch.utils.data.DataLoader(
        mnistm.MNISTM(
            "./mnistm",
            train=False,
            download=True,
            transform=transforms.Compose(
                [
                    transforms.Resize(32),
                    transforms.ToTensor(),
                    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                ]
            ),
        ),
        batch_size=args.test_batch_size,
        shuffle=True,
    )
    svhn_train_loader = torch.utils.data.DataLoader(
        mnistm.MNISTM(
            "./mnistm",
            train=True,
            download=True,
            transform=transforms.Compose(
                [
                    transforms.Resize(32),
                    transforms.ToTensor(),
                    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                ]
            ),
        ),
        batch_size=args.test_batch_size,
        shuffle=True,
    )


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(500, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 500)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x)

model_path = './models/' + args.model

iter_num = int(args.iter)
g21_path = os.path.join(model_path, 'g21-%d.pkl' % (iter_num))

g21 = G21(args, conv_dim=64)

g21.cuda(gpu_ids)

g21.load_state_dict(torch.load(g21_path))

mnist_iter = iter(test_loader)

total_step = len(mnist_iter)

model = Net()
if args.cuda:
    model.cuda(gpu_ids)

optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=args.momentum)


def train(epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(svhn_train_loader):
        if args.cuda:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        # pdb.set_trace()
        optimizer.zero_grad()
        fake_mnist = g21(data)
        # output = model(fake_svhn)
        output = model(fake_mnist)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        '''
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(svhn_train_loader.dataset),
                100. * batch_idx / len(svhn_train_loader), loss.item()))
        '''


def test():
    model.eval()
    test_loss = 0
    correct = 0
    for data, target in svhn_test_loader:
        if args.cuda:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data, volatile=True), Variable(target)
        fake_mnist = g21(data)
        output = model(fake_mnist)
        test_loss += F.nll_loss(output, target, size_average=False).item()  # sum up batch loss
        pred = output.data.max(1, keepdim=True)[1]  # get the index of the max log-probability
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()

    test_loss /= len(svhn_test_loader.dataset)
    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(svhn_test_loader.dataset),
        100. * correct / len(svhn_test_loader.dataset)))


for epoch in range(1, args.epochs + 1):
    train(epoch)
    test()



correct = 0

for i in range(total_step):
    mnist, label = mnist_iter.next()

    mnist = mnist.cuda(gpu_ids)

    mnist = Variable(mnist, volatile=True)
    label = Variable(label)

    # fake_svhn = g21(mnist)
    output = model(mnist)

    pred = output.data.max(1, keepdim=True)[1]  # get the index of the max log-probability
    pred = pred.cpu()
    correct += pred.eq(label.data.view_as(pred)).cpu().sum()

acc = 100.0 * float(correct) / float(total_step)
print('acc: {0}'.format(acc))
